#ifndef ALIGNMENT_HH
#define ALIGNMENT_HH

#include <iostream>
#include "sequence.hh"


/* ============================================================
   class Alignment

   Store a structure-annotated sequence alignment
   
   Methods:
   - append
   for constructing the alignment in the traceback phase
   - write
   for writing the alignment to stdout
   - add_basepairA and add_basepairB
   for adding basepairs in the structures of sequence A and B
   ============================================================ */

class Alignment {
    Sequence &seqA_;
    Sequence &seqB_;
    
    std::vector<int> a_;
    std::vector<int> b_;

    std::vector<char> strA_;
    std::vector<char> strB_;
    
public:
    /**
       Construct Alignment
    
       
    */
    
    Alignment(Sequence &seqA,Sequence &seqB)
	: seqA_(seqA),seqB_(seqB) {
	clear();
    }

    /**
       Delete the alignment edges and reset structure
     */
    void clear() {
	strA_.resize(seqA_.length()+1);
	strB_.resize(seqB_.length()+1);
	for (std::vector<char>::iterator it=strA_.begin(); it!=strA_.end(); ++it) *it='.';
	for (std::vector<char>::iterator it=strB_.begin(); it!=strB_.end(); ++it) *it='.';

	a_.clear();
	b_.clear();
    }

    /**
       Append an alignment edge
    */
    void append(int i, int j) {
	a_.push_back(i);
	b_.push_back(j);
    }

    /**
       Add a basepair to the structure of A
    */
    void add_basepairA(int i, int j) {
	strA_[i]='(';
	strA_[j]=')';
    }

    /**
       Add a basepair to the structure of B
    */
    void add_basepairB(int i, int j) {
	strB_[i]='(';
	strB_[j]=')';
    }

    /**
       Write the alignment to stream out,
       with line-width (without name) width
       If opt_local_out, then output only sequence-locally aligned part
       
       Writes in kind of clustal format without heading line
    */
    void write(std::ostream &out, int width, int score,
	       bool opt_local_out=false, bool opt_pos_out=false) const;

    /**
       Write in clustal format
     */
    void write_clustal(std::ostream &out, int width, int score,
		       bool opt_local_out=false,bool opt_pos_out=false,
		       bool clustal_format=true) const;
   
    /**
       Write in pp format
    */
    void write_pp(std::ostream &out, BasePairs &bpsA, BasePairs &bpsB, int width) const;

    //! get first position of A that is locally aligned to something
    uint get_local_startA() const {return a_[0];}
    
    //! get last position of A that is locally aligned to something
    uint get_local_endA() const {return a_[a_.size()-1];}

    //! get first position of B that is locally aligned to something
    uint get_local_startB()  const {return b_[0];}
    //! get last position of B that is locally aligned to something
    uint get_local_endB() const {return b_[b_.size()-1];}


    //! assignment operator
    const Alignment &operator =(const Alignment &a);

};

#endif
